from __future__ import annotations

from typing import Hashable, List, Optional, Sequence, Tuple, Type, TypeVar

import colorful as cf
import networkx as nx
from jax.tree_util import register_pytree_node

from .annotations import PyTree
from .display import BatchDimensionIterator, display_class, display_generic, display_key_and_value

__all__: List[str] = []

T = TypeVar('T', bound=nx.Graph)


def register_graph_as_jax_pytree(cls: Type[T]) -> None:
    def tree_unflatten(hashed: Hashable, trees: Sequence[PyTree]) -> T:
        node_dicts, edge_dicts = trees

        if not isinstance(node_dicts, dict):
            raise TypeError
        if not isinstance(edge_dicts, dict):
            raise TypeError

        graph = cls()
        graph.add_nodes_from(node_dicts.items())
        graph.add_edges_from([(source, target, data)
                              for (source, target), data in edge_dicts.items()])

        return graph

    def tree_flatten(self: T) -> Tuple[Sequence[PyTree], Hashable]:
        return ((dict(self.nodes), dict(self.edges)), None)

    register_pytree_node(cls, tree_flatten, tree_unflatten)


register_graph_as_jax_pytree(nx.Graph)
register_graph_as_jax_pytree(nx.DiGraph)


@display_generic.register
def _(value: nx.Graph,
      show_values: bool,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    directed = isinstance(value, nx.DiGraph)
    arrow = cf.base00('⟶  ' if directed else '🡘 ')
    retval = display_class(type(value))
    bdi = BatchDimensionIterator(batch_dims)
    for name, node in value.nodes.items():
        sub_batch_dims = bdi.advance(node)
        retval += display_key_and_value(name, node, ": ", show_values, indent + 1, sub_batch_dims)
    for (source, target), edge in value.edges.items():
        key = f"{source}{arrow}{target}"
        sub_batch_dims = bdi.advance(edge)
        retval += display_key_and_value(key, edge, ": ", show_values, indent, sub_batch_dims)
    return retval



from __future__ import annotations

from collections.abc import Mapping
from functools import singledispatch
from numbers import Number
from typing import Any, Iterable, List, Optional, Tuple, Type, TypeVar, Union

import colorful as cf
import numpy as np
from jax.errors import TracerArrayConversionError
from jax.experimental.host_callback import id_tap
from jax.interpreters.ad import JVPTracer
from jax.interpreters.batching import BatchTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer, JaxprTracer
from jax.interpreters.xla import DeviceArray
from jax.tree_util import tree_leaves

from .annotations import Array, TapFunctionTransforms

__all__ = ['print_generic', 'display_generic', 'id_display']


# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/2904.


def print_generic(*args: Any,
                  batch_dims: Optional[Tuple[Optional[int], ...]] = None,
                  raise_on_nan: bool = True,
                  **kwargs: Any) -> None:
    bdi = BatchDimensionIterator(batch_dims)
    found_nan = False
    for value in args:
        sub_batch_dims = bdi.advance(value)
        s = display_generic(value, batch_dims=sub_batch_dims)
        print(s)
        found_nan = found_nan or raise_on_nan and 'nan' in str(s)
    for key, value in kwargs.items():
        sub_batch_dims = bdi.advance(value)
        s = display_key_and_value(key, value, "=", batch_dims=sub_batch_dims)
        print(s)
        found_nan = found_nan or raise_on_nan and 'nan' in str(s)
    if found_nan:
        assert False


@singledispatch
def display_generic(value: Any,
                    show_values: bool = True,
                    indent: int = 0,
                    batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.red(str(value)) + "\n"


@display_generic.register
def _(value: JVPTracer,
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"JVPTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register  # type: ignore
def _(value: JaxprTracer,
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"JaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register  # type: ignore
def _(value: DynamicJaxprTracer,
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"DynamicJaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register  # type: ignore
def _(value: BatchTracer,
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"BatchTracer {value.shape} {value.dtype} "
                      f"batched over {value.val.shape[value.batch_dim]}") + "\n"


@display_generic.register(np.ndarray)  # type: ignore
def _(value: Array,
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    retval = cf.yellow(f"NumPy Array {value.shape} {value.dtype}{_batch_str(batch_dims)}") + "\n"
    return retval + _show_array(indent + 1, value)


@display_generic.register  # type: ignore
def _(value: DeviceArray,
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    base_string = cf.violet(f"Jax Array {value.shape} {value.dtype}") + "\n"
    try:
        np_value = np.asarray(value)
    except TracerArrayConversionError:
        array_string = ""
    else:
        array_string = _show_array(indent + 1, np_value)
    return base_string + array_string


@display_generic.register(type(None))  # type: ignore
@display_generic.register(Number)
def _(value: Union[None, Number],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.cyan(str(value)) + "\n"


@display_generic.register(Mapping)  # type: ignore
def _(value: Mapping[Any, Any],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return (display_class(type(value))
            + "".join(display_key_and_value(key, sub_value, "=", show_values, indent + 1)
                      for key, sub_value in sorted(value.items())))


@display_generic.register(tuple)  # type: ignore
@display_generic.register(list)
def _(value: Union[Tuple[Any, ...], List[Any]],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return display_class(type(value)) + "".join(
        display_key_and_value("", sub_value, "", show_values, indent + 1, sub_batch_dims)
        for sub_batch_dims, sub_value in zip(batch_dimension_iterator(value, batch_dims), value))


_T = TypeVar('_T')


def id_display(x: _T, name: Optional[str] = None, *, no_jvp: bool = False) -> _T:
    def tap(x: _T, transforms: TapFunctionTransforms) -> None:
        nonlocal name
        batch_dims: Optional[Tuple[Optional[int], ...]] = None
        flags = []
        for transform_name, transform_dict in transforms:
            if transform_name == 'batch':
                batch_dims = transform_dict['batch_dims']
                continue
            if no_jvp and transform_name == 'jvp':
                return
            if transform_name in ['jvp', 'mask', 'transpose']:
                flags.append(transform_name)
                continue
        if name is None:
            print_generic(x, batch_dims=batch_dims)
        else:
            if flags:
                final_name = name + f" [{', '.join(flags)}]"
            else:
                final_name = name
            # https://github.com/python/mypy/issues/11583
            print_generic(batch_dims=batch_dims, **{final_name: x})  # type: ignore
    return id_tap(tap, x, result=x)


class BatchDimensionIterator:
    def __init__(self, batch_dims: Optional[Tuple[Optional[int], ...]] = None):
        self.batch_dims = batch_dims
        self.i = 0

    def advance(self, value: Any) -> Optional[Tuple[Optional[int], ...]]:
        if self.batch_dims is None:
            return None
        n = len(tree_leaves(value))  # type: ignore
        old_i = self.i
        self.i += n
        return self.batch_dims[old_i: self.i]

    def check_done(self) -> None:
        if self.batch_dims is not None:
            assert self.i == len(self.batch_dims)


def batch_dimension_iterator(values: Iterable[Any],
                             batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> Iterable[
                                 Optional[Tuple[Optional[int], ...]]]:
    bdi = BatchDimensionIterator(batch_dims)
    for value in values:
        yield bdi.advance(value)
    bdi.check_done()


# Public unexported functions ----------------------------------------------------------------------
def display_class(cls: Type[Any]) -> str:
    return cf.orange(cls.__name__) + "\n"


def display_key_and_value(key: str,
                          value: Any,
                          separator: str,
                          show_values: bool = True,
                          indent: int = 0,
                          batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return (_indent_space(indent) + cf.blue(key) + cf.base00(separator)
            + display_generic(value, show_values, indent, batch_dims))


# Private functions --------------------------------------------------------------------------------
_indentation = 4


def _batch_str(batch_dims: Optional[Tuple[Optional[int], ...]]) -> str:
    if batch_dims is None:
        return ""
    assert len(batch_dims) == 1
    batch_dim = batch_dims[0]
    if batch_dim is None:
        return ""
    return f" batched over axis {batch_dim}"


def _indent_space(indent: int) -> str:
    return (indent * _indentation) * " "


def _format_number(x: float) -> str:
    return f"{x:10.4f}"


def _show_array(indent: int, array: Array) -> str:
    if len(array.shape) == 0:
        return _indent_space(indent) + _format_number(array[()]) + "\n"
    if np.prod(array.shape) == 0:
        return ""
    if 1 in array.shape:
        return _show_array(indent, array[tuple(0 if s == 1 else slice(None)
                                               for s in array.shape)])
    if any(x > 12 for x in array.shape):
        return ""
    if len(array.shape) == 1:
        return _indent_space(indent) + "  ".join(_format_number(array[i])
                                                 for i in range(array.shape[0])) + "\n"
    if len(array.shape) == 2:
        return "".join(_show_array(indent, array_slice)
                       for array_slice in array)
    return ""



from __future__ import annotations

import dataclasses
from dataclasses import MISSING, FrozenInstanceError, InitVar
from functools import partial
from typing import Any, Callable, Hashable, List, Optional, Sequence, Tuple, Type, TypeVar, overload

from jax.tree_util import register_pytree_node

from ..annotations import PyTree
from ..display import BatchDimensionIterator, display_class, display_generic, display_key_and_value
from ..testing import get_relative_test_string, get_test_string, tree_allclose

__all__ = ['dataclass', 'InitVar', 'MISSING', 'FrozenInstanceError']


T = TypeVar('T', bound=Any)


@overload
def dataclass(*, init: bool = True, repr_: bool = True, eq: bool = True,
              order: bool = False) -> Callable[
                  [Type[T]], Type[T]]:
    ...


@overload
def dataclass(cls: Type[T], *, init: bool = True, repr_: bool = True, eq: bool = True,
              order: bool = False) -> Type[T]:
    ...


# TODO: use positional-only arguments
def dataclass(cls: Optional[Type[T]] = None, *, init: bool = True, repr_: bool = True,
              eq: bool = True, order: bool = False) -> Any:
    """
    Returns the same class as was passed in, with dunder methods added based on the fields defined
    in the class.

    Examines PEP 526 annotations to determine fields.  Default values for fields are provided using
    assignment.

    To mark fields as static fields rather than JAX pytree fields, use the `field` function.
    In JAX, a static attribute is one that induces recompilation of a function when it changes, and
    consequently there is more flexibility about what can be done with such an attribute.

    For example::
    ```python
    from __future__ import annotations

    from typing import ClassVar

    import jax.numpy as jnp
    from tjax import Array, dataclass
    from tjax.dataclasses import field
    from jax import grad

    @dataclass
    class LearnedParameter:
        weight: Array
        constrain_positive: bool = field(static=True)
        minimum_positive_weight: ClassVar[Array] = 1e-6

        def trained(self,
                    self_bar: LearnedParameter,
                    learning_rate: float) -> LearnedParameter:
            weight_bar = self_bar.weight
            weight = self.weight - weight_bar * learning_rate
            if self.constrain_positive:
                weight = jnp.maximum(weight, self.minimum_positive_weight)
            return LearnedParameter(weight=weight,
                                    constrain_positive=self.constrain_positive)

    def loss(w: LearnedParameter) -> float:
        return jnp.square(w.weight - 3.3)

    w = LearnedParameter(2.0, True)
    w_bar = grad(loss)(w)
    new_w = w.trained(w_bar, 1e-4)
    ```

    Since this dataclass is a pytree, all of JAX's functions that accept pytrees work with it,
    including iteration, differentiation, and `jax.tree_util` functions.

    Another benefit is the display of dataclasses.  `print(new_w)` gives::
    ```
    LearnedParameter
        weight=Jax Array ()
                2.0003
        constrain_positive=True
    ```
    """
    if cls is None:
        return partial(dataclass, init=init, repr_=repr_, eq=eq, order=order)
    non_none_cls = cls

    # Apply dataclass function to cls.
    data_clz: Type[T] = dataclasses.dataclass(cls, init=init, repr=repr_, eq=eq,
                                              order=order, frozen=True)  # type: ignore

    # Partition fields into hashed, tree, and uninitialized.
    static_fields: List[str] = []
    dynamic_fields: List[str] = []

    for field_info in dataclasses.fields(data_clz):
        if not field_info.init:
            continue
        if field_info.metadata.get('static', False):
            static_fields.append(field_info.name)
        else:
            dynamic_fields.append(field_info.name)

    # Generate additional methods.
    def __str__(self: T) -> str:
        return str(display_generic(self))

    def tree_flatten(x: T) -> Tuple[Sequence[PyTree], Hashable]:
        hashed = tuple(getattr(x, name) for name in static_fields)
        trees = tuple(getattr(x, name) for name in dynamic_fields)
        return trees, hashed

    def tree_unflatten(hashed: Hashable, trees: Sequence[PyTree]) -> T:
        if not isinstance(hashed, tuple):
            raise TypeError
        hashed_args = dict(zip(static_fields, hashed))
        tree_args = dict(zip(dynamic_fields, trees))
        return non_none_cls(**hashed_args, **tree_args)

    # Assign methods to the class.
    if data_clz.__str__ is object.__str__:
        data_clz.__str__ = __str__  # type: ignore

    # Assign field lists to the class.
    data_clz.dynamic_fields = dynamic_fields
    data_clz.static_fields = static_fields

    # Register the class as a JAX PyTree.
    register_pytree_node(data_clz, tree_flatten, tree_unflatten)

    # Register the dynamically-dispatched functions.
    display_generic.register(data_clz, display_dataclass)
    get_test_string.register(data_clz, get_dataclass_test_string)
    get_relative_test_string.register(data_clz, get_relative_dataclass_test_string)
    return data_clz


def display_dataclass(value: T,
                      show_values: bool = True,
                      indent: int = 0,
                      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    retval = display_class(type(value))
    bdi = BatchDimensionIterator(batch_dims)
    for field_info in dataclasses.fields(value):
        sub_value = getattr(value, field_info.name)
        sub_batch_dims = bdi.advance(sub_value)
        retval += display_key_and_value(field_info.name, sub_value, "=", show_values, indent + 1,
                                        sub_batch_dims)
    return retval


def get_dataclass_test_string(actual: Any, rtol: float, atol: float) -> str:
    retval = f"{type(actual).__name__}("
    retval += ",\n".join(
        f"{fn}=" + get_test_string(getattr(actual, fn), rtol, atol)
        for fn in actual.dynamic_fields)
    if actual.dynamic_fields and actual.static_fields:
        retval += ',\n'
    retval += ",\n".join(
        f"{fn}=" + get_test_string(getattr(actual, fn), rtol, atol)
        for fn in actual.static_fields)
    retval += ")"
    return retval


def get_relative_dataclass_test_string(actual: Any,
                                       original_name: str,
                                       original: Any,
                                       rtol: float,
                                       atol: float) -> str:
    retval = f"replace({original_name}, "
    retval += ",\n".join(
        f"{fn}=" + get_relative_test_string(f"{original_name}.{fn}",
                                            getattr(actual, fn),
                                            getattr(original, fn),
                                            rtol,
                                            atol)
        for fn in actual.dynamic_fields
        if not tree_allclose(getattr(actual, fn), getattr(original, fn), rtol=rtol, atol=atol))
    retval += ")"
    return retval



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import jax.numpy as jnp
from jax import vjp
from jax.tree_util import tree_map

from .display import id_display
from .shims import custom_vjp

__all__ = ['copy_cotangent', 'block_cotangent', 'block_variable_cotangent', 'replace_cotangent',
           'print_cotangent', 'CotangentMapper']


X = TypeVar('X')


# copy_cotangent -----------------------------------------------------------------------------------
def copy_cotangent(x: X, y: X) -> Tuple[X, X]:
    return x, y


# Apply after to work around mypy deficiency.
copy_cotangent = custom_vjp(copy_cotangent)  # type: ignore


def _copy_cotangent_fwd(x: X, y: X) -> Tuple[Tuple[X, X], None]:
    return (x, y), None


def _copy_cotangent_bwd(residuals: None, xy_bar: Tuple[X, X]) -> Tuple[X, X]:
    del residuals
    x_bar, _ = xy_bar
    return x_bar, x_bar


copy_cotangent.defvjp(_copy_cotangent_fwd, _copy_cotangent_bwd)  # type: ignore


# replace_cotangent --------------------------------------------------------------------------------
def replace_cotangent(x: X, new_cotangent: X) -> X:
    return x


# Apply after to work around mypy deficiency.
replace_cotangent = custom_vjp(replace_cotangent)  # type: ignore


def _replace_cotangent_fwd(x: X, new_cotangent: X) -> Tuple[X, X]:
    return x, new_cotangent


def _replace_cotangent_bwd(residuals: X, x_bar: X) -> Tuple[X, X]:
    return residuals, x_bar


replace_cotangent.defvjp(_replace_cotangent_fwd, _replace_cotangent_bwd)  # type: ignore


# block_cotangent ----------------------------------------------------------------------------------
_T = TypeVar('_T')


def block_cotangent(f: Callable[..., X],
                    block_argnums: Union[int, Tuple[int, ...]],
                    static_argnums: Union[int, Tuple[int, ...]] = ()) -> Callable[..., X]:
    if isinstance(block_argnums, int):
        block_argnums = (block_argnums,)
    set_block_argnums = set(block_argnums)

    @partial(custom_vjp, static_argnums=static_argnums)
    def blocked_f(*args: Any, **kwargs: Any) -> Any:
        return f(*args, **kwargs)

    def blocked_f_fwd(*args: Any, **kwargs: Any) -> Tuple[Any, Any]:
        return vjp(f, *args, **kwargs)

    def blocked_f_bwd(residuals: Any, output_bar: Any) -> Tuple[Any, ...]:
        f_vjp = residuals
        input_bar = f_vjp(output_bar)
        return tuple(None if i in set_block_argnums else x_bar
                     for i, x_bar in enumerate(input_bar))

    blocked_f.defvjp(blocked_f_fwd, blocked_f_bwd)

    return blocked_f  # type: ignore


@partial(block_cotangent, block_argnums=0)
def block_variable_cotangent(x: _T) -> _T:
    return x


# print_cotangent ----------------------------------------------------------------------------------
def print_cotangent(x: X, name: Optional[str] = None) -> X:
    return x


def _print_cotangent_fwd(x: X, name: Optional[str]) -> Tuple[X, None]:
    return x, None


def _print_cotangent_bwd(name: Optional[str], residuals: None, x_bar: X) -> Tuple[X]:
    del residuals
    x_bar = id_display(x_bar, name)
    return (x_bar,)


# Apply after to work around mypy deficiency.
print_cotangent = custom_vjp(print_cotangent, static_argnums=1)  # type: ignore


print_cotangent.defvjp(_print_cotangent_fwd, _print_cotangent_bwd)  # type: ignore


# print_cotangent ----------------------------------------------------------------------------------
@dataclass
class CotangentMapper:
    cotangent_names: Sequence[str]

    def to_mapping(self, cotangents: _T) -> Mapping[str, _T]:
        """
        Args:
            cotangents: A PyTree whose dynamic fields are arrays representing the group of
                cotangents.
        Returns: A mapping from cotangent names to sub-arrays representing individual cotangents.
        """
        return {cotangent_name: tree_map(lambda x, index=index: x[index], cotangents)
                for index, cotangent_name in enumerate(self.cotangent_names)}

    def vmap(self, x: _T, index: Optional[int]) -> _T:
        """
        Args:
            x: The value of the nonzero cotangent.
            index: The index of the nonzero cotangent.
        Returns: A zeroed copy of x broadcasted to hold each of the cotangents, with up to one
            element holding x.
        """
        def f(x: jnp.ndarray) -> jnp.ndarray:
            z = jnp.zeros((len(self.cotangent_names),) + x.shape, dtype=x.dtype)
            if index is None:
                return z
            return z.at[index].set(x)

        return tree_map(f, x)



from __future__ import annotations

from typing import Any, Mapping, Type

import jax.numpy as jnp
import numpy as np
from jax.dtypes import canonicalize_dtype

__all__ = ['int_dtype',
           'real_dtype',
           'complex_dtype',
           'default_rtol',
           'default_atol',
           'default_tols']


def int_dtype() -> Type[np.signedinteger[Any]]:
    """
    Returns: The type of the dtype used by JAX for int values.  Typically, either `numpy.int32.type`
        or `numpy.int64.type`.
    """
    return canonicalize_dtype(jnp.int_).type


def real_dtype() -> Type[np.floating[Any]]:
    """
    Returns: The type of the dtype used by JAX for int values.  Typically, either
        `numpy.float32.type` or `numpy.float64.type`.
    """
    return canonicalize_dtype(jnp.float_).type


def complex_dtype() -> Type[np.complexfloating[Any, Any]]:
    """
    Returns: The type of the dtype used by JAX for int values.  Typically, either
        `numpy.complex64.type` or `numpy.complex128.type`.
    """
    return canonicalize_dtype(jnp.complex_).type


def default_rtol() -> float:
    return {jnp.float32: 1e-4, jnp.float64: 1e-5}[real_dtype()]


def default_atol() -> float:
    return {jnp.float32: 1e-6, jnp.float64: 1e-8}[real_dtype()]


def default_tols() -> Mapping[str, float]:
    return dict(rtol=default_rtol(), atol=default_atol())



from __future__ import annotations

from typing import Callable, Generic, Optional, Tuple, Type, TypeVar

import jax.numpy as jnp
from jax.tree_util import tree_map, tree_reduce

from ..annotations import PyTree
from ..dataclasses import dataclass
from ..tools import abs_square

__all__ = ['GradientState', 'GradientTransformation', 'SecondOrderGradientTransformation',
           'ThirdOrderGradientTransformation']


# Gradient states ----------------------------------------------------------------------------------
T = TypeVar('T', bound='GenericGradientState')
U = TypeVar('U', bound=PyTree)


@dataclass
class GradientState:
    """
    This base class allows instance checks, and strict type annotations.
    """


@dataclass
class GenericGradientState(GradientState):
    """
    This class serves as a simple container for gradient states because optax doesn't provide a
    suitable base class.
    """
    data: PyTree

    @classmethod
    def wrap(cls: Type[T], weights: Weights, state: PyTree) -> Tuple[Weights, T]:
        return weights, cls(state)


# The weights type variable must be the same type variable that is used for the parameters.  After
# all, these must be compatible pytrees.
Weights = TypeVar('Weights', bound=PyTree)
State = TypeVar('State', bound=GradientState)


@dataclass
class GradientTransformation(Generic[State, Weights]):
    """
    A gradient transformation transforms loss gradients (with respect to weights) into weight deltas
    that are added to the weights.  The typical goal of a gradient transformation is for the weights
    to converge quickly and stably.

    The most basic gradient transformation is the scaled negative gradients, which can be
    interpreted as doing Newton's method with an assumed isotropic loss Hessian.  In general, all
    gradient transformations can be interpreted as Hessian estimators.
    """
    def init(self, parameters: Weights) -> State:
        raise NotImplementedError

    def update(self,
               gradient: Weights,
               state: State,
               parameters: Optional[Weights]) -> Tuple[Weights, State]:
        """
        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.
        """
        raise NotImplementedError


@dataclass
class SecondOrderGradientTransformation(GradientTransformation[State, Weights],
                                        Generic[State, Weights]):
    """
    A second order gradient transformation is a special case of a gradient transformation whose
    update rule accepts a Hessian-vector-product function in its update.
    """
    def update(self,
               gradient: Weights,
               state: State,
               parameters: Optional[Weights]) -> Tuple[Weights, State]:
        """
        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.

        This uses the outer product approximation of the Hessian.
        """
        def hessian_vector_product(v: Weights) -> Weights:
            d = tree_reduce(jnp.add, tree_map(jnp.vdot, gradient, v), 0.0)
            return tree_map(lambda x: x * d, gradient)

        return self.second_order_update(gradient, state, parameters, hessian_vector_product)

    def second_order_update(self,
                            gradient: Weights,
                            state: State,
                            parameters: Optional[Weights],
                            hessian_vector_product: Callable[[Weights], Weights]) -> (
                                Tuple[Weights, State]):
        """
        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
            hessian_vector_product: A function that maps v to the Hessian of the loss with respect
                to the weights times v.
        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.
        """
        raise NotImplementedError


@dataclass
class ThirdOrderGradientTransformation(SecondOrderGradientTransformation[State, Weights],
                                       Generic[State, Weights]):
    """
    A third order gradient transformation is a special case of a second order gradient
    transformation whose update rule accepts the diagonal entries of the Hessian.
    """
    def second_order_update(self,
                            gradient: Weights,
                            state: State,
                            parameters: Optional[Weights],
                            hessian_vector_product: Callable[[Weights], Weights]) -> (
                                Tuple[Weights, State]):
        """
        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
            hessian_vector_product: A function that maps v to the Hessian of the loss with respect
                to the weights times v.
        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.

        Uses the outer product approximation of the Hessian to provide the diagonal entries of the
        Hessian.
        """
        return self.third_order_update(gradient, state, parameters, hessian_vector_product,
                                       tree_map(abs_square, gradient))

    def third_order_update(self,
                           gradient: Weights,
                           state: State,
                           parameters: Optional[Weights],
                           hessian_vector_product: Callable[[Weights], Weights],
                           hessian_diagonal: Weights) -> Tuple[Weights, State]:
        """
        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
            hessian_vector_product: A function that maps v to the Hessian of the loss with respect
                to the weights times v.
            hessian_diagonal: The diagonal entries of the Hessian of the loss with respect to the
                weights.
        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.
        """
        raise NotImplementedError



from typing import Type

import mypy.plugins.dataclasses
from mypy.plugin import Plugin  # pylint: disable=no-name-in-module


class CustomPlugin(Plugin):
    pass


def plugin(version: str) -> Type[CustomPlugin]:
    # pylint: disable=c-extension-no-member
    mypy.plugins.dataclasses.dataclass_makers.add('tjax._src.dataclasses.dataclass.dataclass')
    # mypy.plugins.dataclasses.field_makers.add('tjax._src.dataclasses.helpers.field')
    return CustomPlugin



from __future__ import annotations

from typing import Any, List, Optional, Type, TypeVar, Union

import jax.numpy as jnp
import jax.random
import numpy as np
from jax.random import KeyArray, PRNGKey

from .annotations import Array, BooleanArray, RealArray, RealNumeric, Shape, ShapeLike
from .dataclasses import dataclass

__all__ = ['Generator']


T = TypeVar('T', bound='Generator')
DTypeLikeFloat = Any


@dataclass
class Generator:
    """
    This class represents a JAX random number generator.  Unlike `numpy.Generator`, `tjax.Generator`
    has no mutating methods.  Instead, its generation methods return a new instance along with the
    generated tensor.
    """
    key: KeyArray

    # Class methods --------------------------------------------------------------------------------
    @classmethod
    def from_seed(cls: Type[T], seed: int) -> T:
        return cls(PRNGKey(seed))

    # New methods ----------------------------------------------------------------------------------
    def is_vmapped(self) -> bool:
        return self.key.shape != (2,)

    def split(self, n: int = 2) -> List[Generator]:
        """
        Split a generator into a list of generators.
        """
        assert not self.is_vmapped()
        keys = jax.random.split(self.key, n)
        return [Generator(key=key) for key in keys]

    def vmap_split(self, shape: ShapeLike) -> Generator:
        """
        Split a generator into a generator that can be passed to a vmapped function.
        """
        assert not self.is_vmapped()
        if isinstance(shape, int):
            shape = (shape,)
        else:
            shape = tuple(shape)
        prod_shape = int(np.prod(shape))
        keys = (self.key
                if prod_shape == 1
                else jax.random.split(self.key, prod_shape))
        # https://github.com/google/jax/issues/6473
        keys = keys.reshape(shape + (2,))  # type: ignore
        return Generator(keys)

    def bernoulli(self, p: RealNumeric, shape: Optional[Shape] = None) -> BooleanArray:
        return jax.random.bernoulli(self.key, p, shape)  # type: ignore

    def choice(self,
               a: Union[int, Array],
               shape: Shape = (),
               replace: bool = True,
               p: Optional[RealArray] = None) -> Union[int, Array]:
        return jax.random.choice(self.key, a, shape, replace, p)  # type: ignore

    def gamma(self,
              gamma_shape: RealNumeric,
              shape: Optional[Shape] = None,
              dtype: DTypeLikeFloat = np.float64) -> RealArray:
        return jax.random.gamma(self.key, gamma_shape, shape)  # type: ignore

    def normal(self,
               shape: Shape = (),
               dtype: DTypeLikeFloat = np.float64) -> jnp.ndarray:
        return jax.random.normal(self.key, shape, dtype)

    def uniform(self,
                shape: Shape = (),
                dtype: DTypeLikeFloat = np.float64,
                minval: RealNumeric = 0.,
                maxval: RealNumeric = 1.) -> RealArray:
        return jax.random.uniform(self.key, shape, dtype, minval, maxval)  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt

__all__ = ['Shape', 'ShapeLike', 'SliceLike', 'Array', 'BooleanArray', 'IntegralArray', 'RealArray',
           'ComplexArray', 'Integral', 'Real', 'Complex', 'BooleanNumeric', 'IntegralNumeric',
           'RealNumeric', 'ComplexNumeric', 'PyTree', 'TapFunctionTransforms']

Shape = Tuple[int, ...]
ShapeLike = Union[int, Sequence[int]]
_SliceLikeItem = Union[int, None, slice]
SliceLike = Union[_SliceLikeItem, Tuple[_SliceLikeItem, ...]]


Array = npt.NDArray[Any]
BooleanArray = npt.NDArray[np.bool_]
if TYPE_CHECKING:
    IntegralArray = npt.NDArray[np.integer[Any]]
    RealArray = npt.NDArray[np.floating[Any]]
    ComplexArray = npt.NDArray[Union[np.floating[Any], np.complexfloating[Any, Any]]]
else:
    IntegralArray = npt.NDArray[np.integer]
    RealArray = npt.NDArray[np.floating]
    ComplexArray = npt.NDArray[Union[np.floating, np.complexfloating]]


Integral = int  # Use this until numbers.Integral works with MyPy.
Real = Union[float, int]  # Use this until numbers.Real works with MyPy.
Complex = Union[complex, float, int]  # Use this until numbers.Complex works with MyPy.
BooleanNumeric = Union[BooleanArray, bool]
IntegralNumeric = Union[IntegralArray, Integral]
RealNumeric = Union[RealArray, Real]
ComplexNumeric = Union[ComplexArray, Complex]


PyTree = Any
TapFunctionTransforms = Sequence[Tuple[str, Dict[str, Any]]]



from tjax._src.fixed_point.augmented import AugmentedState
from tjax._src.fixed_point.combinator import (ComparingIteratedFunctionWithCombinator,
                                              IteratedFunctionWithCombinator)
from tjax._src.fixed_point.comparing import ComparingIteratedFunction, ComparingState
from tjax._src.fixed_point.iterated_function import IteratedFunction
from tjax._src.fixed_point.stochastic import (StochasticIteratedFunction,
                                              StochasticIteratedFunctionWithCombinator,
                                              StochasticState)

__all__ = ['AugmentedState', 'ComparingIteratedFunction', 'ComparingIteratedFunctionWithCombinator',
           'ComparingState', 'IteratedFunction', 'IteratedFunctionWithCombinator',
           'StochasticIteratedFunction', 'StochasticIteratedFunctionWithCombinator',
           'StochasticState']



from __future__ import annotations

from numbers import Number
from typing import Any, Optional, Union, overload

import jax.numpy as jnp
import numpy as np

from .annotations import (BooleanNumeric, ComplexArray, ComplexNumeric, IntegralNumeric, RealArray,
                          RealNumeric)

__all__ = ['is_scalar', 'abs_square', 'divide_where', 'divide_nonnegative']


def is_scalar(x: Any) -> bool:
    return isinstance(x, Number) or isinstance(x, (np.ndarray, jnp.ndarray)) and x.shape == ()


@overload
def abs_square(x: ComplexArray) -> RealArray:
    ...


@overload
def abs_square(x: ComplexNumeric) -> RealNumeric:
    ...


def abs_square(x: ComplexNumeric) -> RealNumeric:
    return jnp.square(x.real) + jnp.square(x.imag)


@overload
def divide_where(dividend: RealNumeric,
                 divisor: Union[RealNumeric, IntegralNumeric],
                 *,
                 where: Optional[BooleanNumeric] = None,
                 otherwise: Optional[RealNumeric] = None) -> RealNumeric:
    ...


@overload
def divide_where(dividend: ComplexNumeric,
                 divisor: Union[ComplexNumeric, IntegralNumeric],
                 *,
                 where: Optional[BooleanNumeric] = None,
                 otherwise: Optional[ComplexNumeric] = None) -> ComplexNumeric:
    ...


def divide_where(dividend: ComplexNumeric,
                 divisor: Union[ComplexNumeric, IntegralNumeric],
                 *,
                 where: Optional[BooleanNumeric] = None,
                 otherwise: Optional[ComplexNumeric] = None) -> ComplexNumeric:
    """
    Returns: `jnp.where(where, dividend / divisor, otherwise)`, but without evaluating
        `dividend / divisor` when `where` is false.  This prevents some exceptions.
    """
    if where is None:
        return jnp.true_divide(dividend, divisor)
    dividend = jnp.where(where, dividend, 1.0)
    divisor = jnp.where(where, divisor, 1.0)
    quotient = jnp.true_divide(dividend, divisor)
    return jnp.where(where, quotient, otherwise)


def divide_nonnegative(dividend: RealNumeric, divisor: RealNumeric) -> RealNumeric:
    """
    Returns: The quotient assuming that the dividend and divisor are nonnegative, and infinite
        whenever the divisor equals zero.
    """
    return divide_where(dividend, divisor, where=divisor > 0.0, otherwise=jnp.inf)
